### Copyright  1999 CodeIt Computing, All rights reserved.

## By obtaining, using, and/or copying this software and/or its
## associated documentation, you agree that you have read, understood,
## and will comply with the following terms and conditions:

## Permission to use, copy, modify, and distribute this software and its
## associated documentation for any purpose and without fee is hereby
## granted, provided that the above copyright notice appears in all
## copies, and that both that copyright notice and this permission notice
## appear in supporting documentation, and that the name of CodeIt
## Computing not be used in advertising or publicity pertaining to
## distribution of the software without specific, written prior
## permission.

## CODEIT COMPUTING DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
## SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS. IN NO EVENT SHALL CODEIT COMPUTING BE LIABLE FOR ANY SPECIAL,
## INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
## FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
## NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
## WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

__doc__="""Session Object

WARNING!!! This is a reference implementation only.  I wrote this to
explore session concepts and to get a better idea of what I could or
could not do with Zope's security model.

This code by no means is complete pretty or  even that useful. But it has
proven that the session concept is valid.

What is a session?

A session object is an object that behaves like a folder in zope. If
you access <Session FolderId>/docLogin via the web, the session folder
will validate your username and password based off of the first
acl_users folder that it finds via acquisition. Once the user has been
authorized the Session Folder creates a brand new Session Instance and
stores the user information in the Session Instance. Whenever an
object is accessed through the Session Instance the user is equivalent
to the user that originally logged in. In effect the authorization
information is stored in PATH_INFO instead of a cookie or using
standard HTTP authentication methods. Session Instances will/should be
automatically deleted if the user has been idle longer than the time
to live property.

The advantages to this scheme are. (Of course most of this is not
implemented in this sample)

* A user can be logged in as more than one person at a time, without 
  having to restart there web browser.
  
* An admin can get a list of all the people currently logged into
  Zope.

* You can limit the amount of times the same user id is logged into
  Zope at once.

* The Session Folder should be able to use any type of acl_users
  folder, so authentication can be done against the User, UserDB, or
  UserLDAP already or any other custom User Folder. Because the user
  is only looked up once, Zope does not have to constantly go back
  to the data source for every page request and perform authentication 
  like it does now.

* Sessions can time out values so a user can be logged out if they are
  inactive for to long. i.e. They leave thier computer unattended, they
  will be logged out.

* Sessions can be used to store state information between web requests
  so that the user's current IP address can be stored within the
  session which can prevent people from hijacking sessions. Also makes
  it really easy to implement a shopping cart object. 

* The scheme is token based, which means that the user name and
  password aren't passed around in the clear, so it might be more
  secure. Of course the session would have to be created under SSL for
  maximum protection, but then once the session has been created the
  session could be using normal HTTP for speed.

* There's many more, but I'm tired of typing.

"""




import string, sys, time, whrandom

import OFS.Folder
from OFS.Document import Document
from Globals import HTMLFile, Persistent, PersistentMapping
from Acquisition import Implicit
from AccessControl.User import User, Super, nobody
import anydbm
from sha import sha
import os

from DateTime import DateTime




siad = 86400 # seconds in a day
siah = 3600  # seconds in an hour
siam = 60    # seconds in a minute

addSessionForm=HTMLFile('dtml/addSession', globals())

# TODO; Figure out a better approach to last access time

try:
    DBM = anydbm.open(INSTANCE_HOME + '/accesstimes.dbm', 'c')
except anydbm.error:
##     # DB went bad delete file and try again.
    
    os.unlink(INSTANCE_HOME + '/accesstimes.dbm')
    DBM = anydbm.open(INSTANCE_HOME + '/accesstimes.dbm', 'c')


def addSession(self, id, title="", REQUEST=None):
    """Add a Session to Zope"""
    
    self._setObject(id, Session(id, title))
    
    if REQUEST is not None: 
	# Display the folder that contains the new Session
	return self.manage_main(self,REQUEST)


class BaseSessionInstance(Implicit):
    '''Place Holder'''

    def __init__(self, user, REQUEST):
	self.user = user
        self.loginTime = DateTime()
	#self._v_accessedTime = time.time()
	id = "%s:%s:%s" % (self.loginTime,
			   user.getUserName(), 
			   REQUEST.REMOTE_ADDR)
	id  = sha(id)
	self.id = '$' + id.hexdigest()[:10]
        
	# This object will preform it's own validation
	self.__allow_groups__ = self
	self.updateLastAccessedTime()

        self.brandSession(REQUEST)

        

    def __str__(self): return self.user.getUserName()


    def brandSession(self, REQUEST):
        # Set a cookie to prevent hijacking
        brands = REQUEST.cookies.get('session-brands', None)

        if brands is not None:
            brands = string.split(brands, ':')
        else:
            brands = []
        
        brand = hex(whrandom.randint(0,sys.maxint-1))[2:]
        self.__brand__ = brand

        brands.append(brand)
        brands = string.join(brands, ':')
        
        resp=REQUEST['RESPONSE']
        resp.setCookie('session-brands', brands, path='/')

    def validBrand(self, REQUEST):
        brands = REQUEST.cookies.get('session-brands', None)

        if brands is not None:
            brands = string.split(brands, ':')
        else:
            brands = []

        return self.__brand__ in brands

    
    def CurrentSession(self):
	return self

    
    def updateLastAccessedTime(self):
	DBM[self.id] = str(time.time())

    def getLastAccessedTime(self):
	try:
	    return float(DBM[self.id])
	except KeyError:
	    # For some reason the value is not in the DBM file
	    self.updateLastAccessedTime()
	    return time.time()
	    
	

    def idle(self):
	#idle = time.time() - self._v_accessedTime
	idle = time.time() - self.getLastAccessedTime()
        return idle

    def strIdle(self): 
        days, rem = divmod(int(self.idle()), siad)
        hours, rem = divmod(rem, siah)
        minutes, seconds = divmod(rem, siam)
    
        if days == 1:
            fmts = "%s day %s:%s:%s"
        else:
            fmts =  "%s days %s:%s:%s"

        return fmts % (days, hours, minutes, seconds)

    
    def validate(self,REQUEST,auth='',roles=None): 
	# Try to authorize user
	if self.user.allowed(self, roles):
	    return self.user
	else:
	    # The login does not have the appropriate permissions.
	    # But first figure out what the destination was so we can
	    # get back to the request afterwards.
	    # What about qureis?

	    sessionURL = "%s/%s/" % (self.absolute_url(), self.id)
	    destination = REQUEST.SERVER_URL + REQUEST.PATH_INFO
	    destination = destination[len(sessionURL):]
	    raise 'LoginRequired', self.docLogin(self, REQUEST, destination=destination)

    def set(self, attr, value):
	# Set random information in the session. 
	# Note: We definatly need some security/protection mechanisms
	# to keep people using dtml from overwritting import info. i.e.
	# "user"

        objs = string.split(attr, '.')
 
        obj = self
        attr = objs[0]

	# Walk the object's sub objects if necessary
        for attr in objs[:-1]:
            obj = getattr(obj, attr)

	attr = objs[-1]
        
        setattr(obj, attr, value)

class SessionInstance(BaseSessionInstance, Persistent):
    "ZODB Based Session Instance"
    pass


from sqlSessionStorage import sqlSessionStorage, sqlSessionInstance

class Session(
    OFS.Folder.Folder
    ):
    """A Session
    """
    
    icon='misc_/Session/icon'   
    meta_type='Session'	# Objects Name in the Zope list

    manage_sessions=HTMLFile('dtml/Sessions', globals())

    manage_options = (OFS.Folder.Folder.manage_options[0],) +\
                     ({'label':'Sessions',   'action':'manage_sessions'},) +\
                     OFS.Folder.Folder.manage_options[1:]

    # There should be a way to select this.
    # Instance = sqlSessionInstance
    Instance = SessionInstance

    def __init__(self, id, title=''):
	self.id = id
	self.title = title
	self.sessions = PersistentMapping()
        
        #self.sessions = sqlSessionStorage()

	self._setObject('docLogin', Document(_docLogin, __name__='docLogin'))
	self._setObject('docLogout', Document(_docLogout, __name__='docLogout'))
	self._init()

    def _init(self, *args):
	pass
	
    def __bobo_traverse__(self, REQUEST, name=None):
	if name[0] !="$":
	    return getattr(self, name)
	elif self.sessions.has_key(name):
            session = self.sessions[name]

            if not session.validBrand(REQUEST):
                # Uh oh... the user's cookie does not match. We should
                # make them log in.

                return PathEater('%s/%s' % (REQUEST.URL1, 'docLogin'))
            
	    session.updateLastAccessedTime()
            return session.__of__(self)
	else: 
	    # Eat the path and redirect to the login page.

	    return PathEater('%s/%s' % (REQUEST.URL1, 'docLogin'))


    def anonymousLogin(self, REQUEST, RESPONSE, destination="index_html"):
	"Create an Anonymous session" 
	newSession = self.Instance(nobody, REQUEST)
	self.sessions[newSession.id] = newSession
	destination = "%s/%s/%s" % (REQUEST.URL1, newSession.id, destination)
	return RESPONSE.redirect(destination)

    def login(self, username, password, REQUEST, 
		     RESPONSE, destination=""):
	"Login the user and create a new session"

	
	acl = getattr(self, 'acl_users')
	super = acl._super
	# Check for superuser

	# Note: This code was lifted from User.py and slightly modified.
 	# It dosen't appear to let the superuser log-into a session.
	# This of course may be a good or bad thing, if you think about it.
	if acl._isTop() and (username==super.getUserName()):
	    user = super
	else:
	    # Try to get user
	    user=acl.getUser(username)
	
	if (user is not None) and  user.authenticate(password, REQUEST):
	    # We need the user to be able to acquire!
	    # Not sure why --Scott
	    user=user.__of__(self)
	    
            #TODO: Session ID's need to random
	    
	    newSession = self.Instance(user, REQUEST)
	    self.sessions[newSession.id] = newSession
	    destination = "%s/%s/%s" % (REQUEST.URL1, newSession.id, destination)
	    return RESPONSE.redirect(destination)
	else:
	    raise 'LoginRequired', self.docLogin(self, REQUEST)
	    

    def manage_bootSessions(self, ids=[], REQUEST=None):
	"""Remove sessions by hand"""

	for id in ids:
            try:
                del self.sessions[id]
	    except:
		pass
	    
	if REQUEST is not None:
	    return self.manage_sessions(self, REQUEST)

    def manage_bootAllSessions(self, idle=0, REQUEST=None):
	"""Remove all sessions"""
        idle = int(idle) 

        if idle == 0:
            self.sessions.clear()
        else:
            sessions = self.sessions.values()
            for session in sessions:
                if session.idle() > idle:
                    del self.sessions[session.id]

	if REQUEST is not None:
	    return self.manage_sessions(self, REQUEST)


class PathEater:
    "Eats path info for breakfast"
    
    def __init__(self, url):
	self.destination = ""
	self.url = url
    def __bobo_traverse__(self, request, name):
	self.destination = self.destination + "/%s" % name
	return self
          
    def __call__(self, REQUEST, RESPONSE):
	"Publish method"
	return RESPONSE.redirect("%s?destination=%s" % (self.url, self.destination))
	#return "I ate %s" % self.steps




    


_docLogin='''<HTML>
<HEAD>
<TITLE>Login</TITLE>
</HEAD>
<BODY BGCOLOR="#FFFFFF" LINK="#000099" VLINK="#555555">
<P>
<FORM ACTION="<!--#var URL1-->/login" METHOD="POST">

<!--#if destination-->
<input type=hidden name="destination" value="<!--#var destination-->">
<!--#/if-->

<TABLE>
<TR>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <STRONG>Name</STRONG>
  </TD>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <INPUT TYPE="TEXT" NAME="username" SIZE="20">
  </TD>
</TR>

<TR>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <STRONG>Password</STRONG>
  </TD>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <INPUT TYPE="PASSWORD" NAME="password" SIZE="20">
  </TD>
</TR>
<TR>
  <TD ALIGN="LEFT" VALIGN="TOP">
  </TD>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <INPUT TYPE="SUBMIT" NAME="submit" VALUE=" Ok ">
  </TD>
</TR>
</TABLE>
</FORM>
</BODY>
</HTML>'''


_docLogout='''<HTML>
<HEAD>
<TITLE>Logout</TITLE>
</HEAD>
<BODY BGCOLOR="#FFFFFF" LINK="#000099" VLINK="#555555">
<P>
<CENTER>
You have been logged out of the system.
</CENTER>
</BODY>
</HTML>'''
